

class RWSDeque {   /* https://stackoverflow.com/a/60055110, simplified for our usage */
    constructor() {
        this.front = this.back = undefined;
    }
    addFront(value) {
        if (!this.front) this.front = this.back = { value };
        else this.front = this.front.next = { value };
    }
    removeBack() {
        let value = this.peekBack();
        if (this.front === this.back) this.front = this.back = undefined;
        else (this.back = this.back.next).back = undefined;
        return value;
    }
    peekBack() {
        return this.back && this.back.value;
    }
    forEach(callback) {   /* in back-first, front-last order */
        var cur = this.back;
        while (cur) {
            callback(cur.value);
            cur = cur.next;
        }
    }
}

function RetryWebSocket(url) {
    const VERSION_INFO = "1";
    const TIMEOUT = 10;    /* seconds, approximately */
    const rws = {
        url: url,
        onopen: null,
        onmessage: null,
        onclose: null,
    };
    const rws_url = url + (url.includes("?") ? "&" : "?") + "rws=";

    /* status:
     * 0 = attempting initial connection
     * 1 = connected normally
     * 2 = attempting reconnection
     * -1 = closed
     */
    var status = 0;
    var retry_id = VERSION_INFO, timer_id, timer_remaining;
    var reconnect_attempts, max_reconnect_attempts;
    var out_base, out_queue;
    var in_recv = 0;
    var current_ws = null;

    function close_definitely(msg) {
        if (status >= 0) {
            if (current_ws !== null) {
                current_ws.send("@-1");
            }
            status = -1;
            console.log("RetryWebSocket: " + msg);
            close_ws();
            out_queue = null;

            if (rws.onclose !== null)
                rws.onclose();
        }
    }

    function do_ontimer() {
        if (timer_id === null)
            return;
        if (status <= 0) {
            clearInterval(timer_id);
            timer_id = null;
            return;
        }
        timer_remaining -= 2;
        if (timer_remaining <= 0) {
            timer_remaining = TIMEOUT;
            close_ws();
            if (status == 1) {
                status = 2;
                reconnect_attempts = 1;
            }
            else {
                reconnect_attempts += 1;
                if (reconnect_attempts > max_reconnect_attempts) {
                    close_definitely("give up auto-reconnecting after " + max_reconnect_attempts + " attempts");
                    return;
                }
            }
            open_ws();
        }
        else if (status == 1) {
            /* works as a keepalive, too, so we send it even in case 'in_recv' didn't change */
            current_ws.send("@" + in_recv);
        }
    }

    function close_ws() {
        const ws = current_ws;
        if (ws !== null) {
            current_ws = null;
            ws.close();
        }
    }

    function got_ack(n) {
        if (n < out_base) {
            close_definitely("bogus ACK (too small)");
            return;
        }
        while (n > out_base) {
            if (out_queue.removeBack() === undefined) {
                close_definitely("bogus ACK (too large)");
                return;
            }
            out_base += 1;
        }
    }

    function open_ws() {
        const ws = new WebSocket(rws_url + retry_id);
        ws.binaryType = "arraybuffer";
        console.log("connecting to " + ws.url);
        current_ws = ws;

        ws.onclose = function (event) {
            if (current_ws === ws) {
                current_ws = null;
                if (status == 1) {
                    status = 2;
                    timer_remaining = 0;   /* try to reconnect quickly */
                    reconnect_attempts = 0;
                }
                else if (status <= 0) {
                    close_definitely("connection lost before it was ready");
                    return;
                }
                else {
                    console.log("reconnecting failed, will continue to try");
                }
            }
        }

        ws.onmessage = function (event) {
            if (current_ws === ws) {
                timer_remaining = TIMEOUT;
                if (typeof event.data == "string" && event.data[0] == "@") {
                    const cmd = JSON.parse(event.data.substring(1));
                    if (typeof (cmd) == "number") {
                        got_ack(cmd);
                    }
                    else {
                        console.log(event.data);
                        switch (cmd["cmd"]) {

                            case "hello":
                                if (status != 0) {
                                    close_definitely("unexpected 'hello'");
                                    return;
                                }
                                retry_id = cmd["rws"];
                                max_reconnect_attempts = cmd["gtimeout"] / TIMEOUT;
                                out_base = 0;
                                out_queue = new RWSDeque;
                                in_recv = 0;
                                timer_id = setInterval(do_ontimer, 2000);
                                timer_remaining = TIMEOUT;
                                status = 1;

                                if (rws.onopen !== null)
                                    rws.onopen();
                                break;

                            case "reconnect":
                                if (status != 2) {
                                    close_definitely("unexpected 'reconnect'");
                                    return;
                                }
                                got_ack(cmd["ack"]);
                                ws.send("@" + in_recv);
                                out_queue.forEach(msg => ws.send(msg));
                                timer_remaining = TIMEOUT;
                                status = 1;
                                break;
                                
                            case "close":
                                close_definitely("server-side close (" + cmd["msg"] + ")");
                                break;
                        }
                    }
                }
                else {
                    in_recv += 1;
                    if (rws.onmessage !== null)
                        rws.onmessage(event);
                }
            }
        }
    }

    function send(msg) {
        if (status == 1) {
            current_ws.send(msg);
        }
        else if (status < 1) {
            if (status == 0)
                close_definitely("send() called too early");
            return;
        }
        out_queue.addFront(msg);
    }
    rws.send = send;

    function close() {
        close_definitely("close()");
    }
    rws.close = close;

    open_ws();
    return rws;
}


function encode_arraybuffer_to_base64(buffer) {
    var binary = '';
    var bytes = new Uint8Array(buffer);
    var len = bytes.byteLength;
    for (var i = 0; i < len; i++) {
        binary += String.fromCharCode(bytes[i]);
    }
    return btoa(binary);
}

function decode_base64_to_arraybuffer(base64) {
    var raw = window.atob(base64);
    var rawLength = raw.length;
    var buffer = new ArrayBuffer(rawLength);
    var array = new Uint8Array(buffer);

    for (i = 0; i < rawLength; i++) {
        array[i] = raw.charCodeAt(i);
    }
    return buffer;
}
